<?php
/**
 * Creates extra api endpoints for the plugin.
 *
 * @package Fast Plugin
 *
 * @copyright 2017 Oxygenna.com
 * @license http://themeforest.net/licenses/standard
 * @version 1.5.0
 * @author Oxygenna
 */

/**
 * Adds extra endpoints for api
 */
class FastAPI extends WP_REST_Controller
{
    /**
     * Constructor for the api.
     */
    public function __construct()
    {
        $this->namespace     = '/fast-plugin/v1';
    }

    /**
     * Registers routes with WordPress.
     */
    public function register_routes()
    {
        register_rest_route($this->namespace, '/is_logged_in', array(
            array(
                'methods'  => 'GET',
                'callback' => array($this, 'route_is_logged_in'),
                'args'     => array()
            ),
        ));

        register_rest_route($this->namespace, '/envato_login', array(
            array(
                'methods'   => 'POST',
                'callback'  => array($this, 'route_envato_login'),
                'args' => array(
                    'code' => array(
                        'required' => true
                    ),
                    'nonce' => array(
                        'required' => true
                    )
                )
            ),
        ));

        register_rest_route($this->namespace, '/register', array(
            array(
                'methods'   => 'POST',
                'callback'  => array($this, 'route_register'),
                'args' => array(
                    'first_name' => array(
                        'required' => true
                    ),
                    'last_name' => array(
                        'required' => true
                    ),
                    'email' => array(
                        'required' => true
                    ),
                    'password' => array(
                        'required' => true
                    ),
                    'captcha' => array(
                        'required' => false
                    ),
                    'nonce' => array(
                        'required' => true
                    ),
                    'language' => array(
                        'required' => false
                    )
                )
            )
        ));

        register_rest_route($this->namespace, '/login', array(
            array(
                'methods'   => 'POST',
                'callback'  => array($this, 'route_login'),
                'args' => array(
                    'email' => array(
                        'required' => true
                    ),
                    'password' => array(
                        'required' => true
                    ),
                    'nonce' => array(
                        'required' => true
                    )
                )
            )
        ));

        register_rest_route($this->namespace, '/current_user', array(
            array(
                'methods'             => 'GET',
                'callback'            => array($this, 'route_current_user'),
                'permission_callback' => array($this, 'user_logged_in_permission_check'),
                'args'                => array()
            ),
        ));

        register_rest_route($this->namespace, '/logout', array(
            array(
                'methods'             => 'GET',
                'permission_callback' => array($this, 'user_logged_in_permission_check'),
                'callback'            => array($this, 'route_logout')
            )
        ));

        register_rest_route($this->namespace, '/extra_fields', array(
            array(
                'methods'             => 'GET,POST',
                'callback'            => array($this, 'route_extra_fields'),
                'permission_callback' => array($this, 'user_logged_in_permission_check'),
            )
        ));
    }

    /**
     * Used for permission_callback on routes
     *
     * @return boolean If user is logged in.
     */
    public function user_logged_in_permission_check()
    {
        return is_user_logged_in();
    }

    /**
     * Checks if a user is logged in now.
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_is_logged_in($request)
    {
        return is_user_logged_in();
    }

    /**
     * Returns current user
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_current_user($request)
    {
        return $this->get_user_data();
    }

    /**
     * Logs into envato api using code provided by login page param.
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_envato_login($request)
    {
        if (false === wp_verify_nonce($request->get_param('nonce'), 'ticket-api-action')) {
            return new WP_Error('cant-login', esc_html__('Invalid Nonce', 'fast-plugin'), array(
                'status' => 401
            ));
        }

        // Get envato token.
        $token_response = wp_remote_post('https://api.envato.com/token/', array(
            'method' => 'POST',
            'timeout' => 30,
            'headers' => array(
                'Content-type' => 'application/x-www-form-urlencoded'
            ),
            'body' => array(
                'grant_type'    => 'authorization_code',
                'code'          => $request->get_param('code'),
                'client_id'     => fast_get_option('envato_client_id'),
                'client_secret' => fast_get_option('envato_client_secret')
            )
        ));

        if (is_wp_error($token_response)) {
            return $token_response;
        }

        // Decode the token.
        $token = json_decode(wp_unslash(wp_remote_retrieve_body($token_response)));

        // Check for error responses.
        if (200 !== wp_remote_retrieve_response_code($token_response)) {
            return new WP_Error($token->error, $token->error_description);
        }

        // Use token to get the user details.
        $user_details = $this->envato_get_user($token->access_token);

        if (is_wp_error($user_details)) {
            return $user_details;
        }

        // If user exists log in, otherwise register user.
        $user = get_user_by('email', $user_details['email']);
        if (false === $user) {
            // No user so we need to register one.
            $user = fast_register_user(
                $user_details['account']['firstname'],
                $user_details['account']['firstname'],
                $user_details['account']['surname'],
                $user_details['email'],
                wp_generate_password()
            );

            if (is_wp_error($user)) {
                return $user;
            }
        }

        $purchases = $this->envato_get($token->access_token, 'https://api.envato.com/v3/market/buyer/purchases');

        // Check for errors.
        if (is_wp_error($purchases)) {
            return $purchases;
        }

        // Create user purchases.
        if (isset($purchases['purchases']) && is_array($purchases['purchases'])) {
            // Allow empty content & title for posts.
            add_filter('wp_insert_post_empty_content', '__return_false');

            foreach ($purchases['purchases'] as $purchase) {
                $this->add_envato_purchase($purchase, $user, $user_details);
            }
        }

        return $this->login_user($user);
    }

    /**
     * Adds a purchase to the users purchase list.
     *
     * @param array $envato_purchase Purchase data to add.
     * @param array $wp_user WordPress user.
     * @param array $envato_user  Envato user data.
     *
     * @return array New list of purchases
     */
    private function add_envato_purchase($envato_purchase, $wp_user, $envato_user)
    {
        $purchase = fast_get_user_envato_purchase($envato_purchase['code'], $wp_user->ID);
        if (false === $purchase) {
            // This purchase is new.
            $purchase = array(
                'post_status' => 'private',
                'post_author' => $wp_user->ID,
                'post_type'   => 'fast_purchase'
            );
        } else {
            // Convert the WP_Post to array.
            $purchase = $purchase->to_array();
        }

        // Create purchase meta data.
        $purchase['meta_input'] = array(
            '_buyer'           => $envato_user['username'],
            '_created_at'      => $envato_purchase['sold_at'],
            '_item_id'         => $envato_purchase['item']['id'],
            '_item_name'       => $envato_purchase['item']['name'],
            '_licence'         => $envato_purchase['license'],
            '_purchase_code'   => $envato_purchase['code'],
            '_supported_until' => $envato_purchase['supported_until'],
        );

        // Now look up the product taxonomy for this product so we can link it.
        $products = get_terms(array(
            'hide_empty' => false,
            'taxonomy'   => 'fast_product',
            'meta_key'   => '_fast-plugin-product_envato_item',
            'meta_value' => $envato_purchase['item']['id'],
            'fields'     => 'ids'
        ));

        if (count($products) > 0) {
            $purchase['meta_input']['_ticket_product'] = (int) $products[0];
        } else {
            // Product not supported so return.
            return;
        }

        return wp_insert_post($purchase, true);
    }

    /**
     * Fetches user info from Envato API.
     *
     * @param string $token Bearer token.
     *
     * @return array User information.
     */
    public function envato_get_user($token)
    {
        // Get account details.
        $account = $this->envato_get($token, 'https://api.envato.com/v1/market/private/user/account.json');

        if (is_wp_error($account)) {
            return $account;
        }

        $email = $this->envato_get($token, 'https://api.envato.com/v1/market/private/user/email.json');

        if (is_wp_error($email)) {
            return $email;
        }

        $username = $this->envato_get($token, 'https://api.envato.com/v1/market/private/user/username.json');

        if (is_wp_error($username)) {
            return $username;
        }

        return array_merge($account, $email, $username);
    }

    /**
     * Fetches info from Envato API.
     *
     * @param string $token Bearer token.
     * @param string $url to call.
     *
     * @return object Response from get or WP_Error.
     */
    private function envato_get($token, $url)
    {
        $response = wp_remote_get($url, array(
            'timeout' => 30,
            'headers' => array(
                'Authorization' => 'Bearer ' . $token
            )
        ));

        if (is_wp_error($response)) {
            return $response;
        }

        $value = json_decode(wp_remote_retrieve_body($response), true);

        // Check for error responses.
        if (200 !== wp_remote_retrieve_response_code($response)) {
            return new WP_Error($value->error, $value->error_description);
        }

        return $value;
    }

    /**
     * Logs a user in to WordPress
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_login($request)
    {
        if ($request->get_header('referer') !== trailingslashit(get_the_permalink(fast_get_option('support_page'))) . 'login') {
            return new WP_Error('cant-login', esc_html__('Must log in from the app', 'fast-plugin'), array(
                'status' => 401
            ));
        }

        if (false === wp_verify_nonce($request->get_param('nonce'), 'ticket-api-action')) {
            return new WP_Error('cant-login', esc_html__('Invalid Nonce', 'fast-plugin'), array(
                'status' => 401
            ));
        }

        $user = wp_authenticate($request->get_param('email'), $request->get_param('password'));
        return $this->login_user($user);
    }

    /**
     * Logs a user into WordPress.
     *
     * @param object $user User object.
     *
     * @return object|array User data in array or WP_Error object.
     */
    private function login_user($user)
    {
        if (!is_wp_error($user)) {
            // Login user.
            wp_set_current_user($user->ID, $user->email);
            wp_set_auth_cookie($user->ID);
            do_action('wp_login', $user->ID);

            $cookie = wp_parse_auth_cookie('', 'logged_in');

            // We have a new nonce now cookies have changed make sure we return it in the headers.
            global $wp_rest_server;
            $wp_rest_server->send_header('X-WP-Nonce', wp_create_nonce('wp_rest'));

            // return user data.
            return $this->get_user_data();
        }

        return new WP_Error('cant-login', esc_html__('Invalid username or password', 'fast-plugin'), array(
            'status' => 401
        ));
    }

    /**
     * Registers a user in to WordPress
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_register($request)
    {
        if ($request->get_header('referer') !== trailingslashit(get_the_permalink(fast_get_option('support_page'))) . 'register') {
            return new WP_Error('cant-login', esc_html__('Must register in the app', 'fast-plugin'), array(
                'status' => 401
            ));
        }

        if (false === wp_verify_nonce($request->get_param('nonce'), 'ticket-api-action')) {
            return new WP_Error('cant-login', esc_html__('Invalid Nonce', 'fast-plugin'), array(
                'status' => 401
            ));
        }

        $recaptcha_site_key = fast_get_option('recaptcha_site_key');
        $recaptcha_secret_key = fast_get_option('recaptcha_secret_key');

        if (!empty($recaptcha_site_key) && !empty($recaptcha_secret_key)) {
            $response = wp_remote_post('https://www.google.com/recaptcha/api/siteverify', array(
                'method' => 'POST',
                'blocking' => true,
                'timeout' => 60,
                'body' => array(
                    'secret' => $recaptcha_secret_key,
                    'response' => $request->get_param('captcha')
                )
            ));

            if (is_wp_error($response)) {
                return $response;
            }

            $response = json_decode($response['body'], true);
            if (true !== $response['success']) {
                return new WP_Error('invalid-captcha', $response['error-codes'], array(
                    'status' => 401
                ));
            }
        }

        $language = $request->get_param('language');

        $user = fast_register_user(
            $request->get_param('first_name'),
            $request->get_param('first_name'),
            $request->get_param('last_name'),
            $request->get_param('email'),
            $request->get_param('password'),
            null === $language ? '' : $language
        );

        if (is_wp_error($user)) {
            return $user;
        } else {
            update_user_meta($user->ID, '_purchases', array());
            return $this->login_user($user);
        }
    }

    /**
     * Logs a user out of WordPress
     *
     * @param object $request WP_REST_Request object.
     */
    public function route_logout($request)
    {
        if (!is_user_logged_in()) {
            return new WP_Error('cant-logout', esc_html__('User not logged in', 'fast-plugin'), array(
                'status' => 404
            ));
        }

        // Logout user.
        wp_logout();

        $cookie = wp_parse_auth_cookie('', 'logged_in');

        // We have a new nonce now cookies have changed make sure we return it in the headers.
        global $wp_rest_server;
        $wp_rest_server->send_header('X-WP-Nonce', wp_create_nonce('wp_rest'));

        return array(
            'message' => 'Successfully logged out',
        );
    }

    /**
     * Gets and sets the extra fields option.
     *
     * @param object $request HTTP Request.
     */
    public function route_extra_fields($request)
    {
        switch ($request->get_method()) {
            case 'GET':
                // do nothing but return the fields.
                break;
            case 'POST':
                $extra_fields = $request->get_body();
                update_option('fast_extra_fields', json_decode($extra_fields));
                break;
        }
        return fast_get_extra_fields();
    }

    /**
     * Returns user data that we need for the app
     */
    private function get_user_data()
    {
        global $post;
        $current_user = wp_get_current_user();

        return array(
            'id'                => $current_user->ID,
            'first_name'        => $current_user->first_name,
            'last_name'         => $current_user->last_name,
            'description'       => $current_user->description,
            'avatar'            => get_avatar_url($current_user->user_email),
            'is_agent'          => fast_is_user_agent($current_user->ID),
            'ticket_count'      => (int) get_user_meta($current_user->ID, '_ticket_count', true),
            'purchases'         => get_user_meta($current_user->ID, '_purchases', true),
            'profile_edit_url'  => get_edit_user_link(),
            'language'          => get_user_locale()
        );
    }
}
