<?php
/**
 * Functions for the plugin
 *
 * @package Fast Plugin
 *
 * @copyright 2017 Oxygenna.com
 * @license http://themeforest.net/licenses/standard
 * @version 1.5.0
 * @author Oxygenna
 */

/**
 * Creates the default roles when plugin is activated.
 **/
function fast_create_plugin_default_roles()
{
    $roles = apply_filters('fast_filter_ticket_roles', array());
    foreach ($roles as $role => $name) {
        $defaults = apply_filters('fast_filter_default_' . $role . '_capabilities', array());
        $capabilities = array();
        foreach ($defaults as $default) {
            $capabilities[$default] = true;
        }
        add_role($role, $name, $capabilities);
    }
}

/**
 * Checks if a user is a ticket agent.
 *
 * @param int $user_id Id of user to check.
 */
function fast_is_user_agent($user_id)
{
    $agents_with_user_id = get_terms('fast_agent', array(
        'hide_empty' => false,
        'meta_query' => array(
            array(
                'key'       => '_fast-plugin-agent_user_id',
                'value'     => $user_id,
            )
        )
    ));
    return !empty($agents_with_user_id) === true || is_wp_error($agents_with_user_id) === true;
}

/**
 * Gets a user object of an agent from a ticket.
 *
 * @param integer $ticket_id Ticket ID to get the agent from.
 *
 * @return object|boolean User object or false if not found.
 */
function fast_get_ticket_agent($ticket_id)
{
    $agent_terms = wp_get_post_terms($ticket_id, 'fast_agent');
    if (is_wp_error($agent_terms) || empty($agent_terms)) {
        return false;
    }
    $agent_term_id = $agent_terms[0]->term_id;
    $agent_id = fast_get_taxonomy_metabox_value('agent_user_id', $agent_term_id);
    return get_userdata($agent_id);
}

/**
 * Creates an email content subject sends it to the template and then sends the email
 *
 * @param string $sender User who is making the change.
 * @param string $receiver User to send the mail to.
 * @param string $subject Email subject.
 * @param string $message Email message with placeholders.
 * @param array  $ticket  Ticket data.
 * @param object $comment Comment object.
 *
 * @return boolean Whether the email contents were sent successfully.
 */
function fast_send_ticket_email($sender, $receiver, $subject, $message, $ticket, $comment)
{
    $subject                            = apply_filters('fast_filter_replace_placeholders', $subject, $ticket, $comment, $sender, $receiver);
    $message                            = apply_filters('fast_filter_replace_placeholders', $message, $ticket, $comment, $sender, $receiver);
    $ticket_url                         = get_permalink($ticket->ID);
    $ticket_email_template_bg_color     = fast_get_option('ticket_email_template_bg_color');
    $ticket_email_logo_image            = fast_get_option('ticket_email_logo_image');
    $ticket_email_logo_image_bg_color   = fast_get_option('ticket_email_logo_image_bg_color');
    $ticket_email_logo_title            = fast_get_option('ticket_email_logo_title');
    $ticket_email_logo_title_color      = fast_get_option('ticket_email_logo_title_color');
    $ticket_email_logo_title_bg_color   = fast_get_option('ticket_email_logo_title_bg_color');
    $ticket_email_divider_color         = fast_get_option('ticket_email_divider_color');
    $ticket_email_button_color          = fast_get_option('ticket_email_button_color');
    $ticket_email_button_bg_color       = fast_get_option('ticket_email_button_bg_color');
    $ticket_email_above_button          = fast_get_option('ticket_above_ticket_button');
    $ticket_email_footer                = fast_get_option('ticket_email_footer');
    $ticket_email_border_radius         = fast_get_option('ticket_email_border_radius');
    $ticket_email_footer_color          = fast_get_option('ticket_email_footer_color');

    ob_start();
    include apply_filters('fast_filter_email_template', FAST_PLUGIN_DIR . 'templates/email/template-ticket-email.php');
    $email = ob_get_contents();
    ob_end_clean();

    return wp_mail($receiver->user_email, $subject, $email, array(
        'Content-Type: text/html; charset="UTF-8"',
    ));
}

/**
 * Sends new user email when user registers
 *
 * @param string $to Email address to send to.
 * @param string $subject Subject of the email.
 * @param string $message Message to send in the email.
 **/
function fast_send_regular_email($to, $subject, $message)
{
    $ticket_email_template_bg_color   = fast_get_option('ticket_email_template_bg_color');
    $ticket_email_logo_image          = fast_get_option('ticket_email_logo_image');
    $ticket_email_logo_image_bg_color = fast_get_option('ticket_email_logo_image_bg_color');
    $ticket_email_logo_title          = fast_get_option('ticket_email_logo_title');
    $ticket_email_logo_title_color    = fast_get_option('ticket_email_logo_title_color');
    $ticket_email_logo_title_bg_color = fast_get_option('ticket_email_logo_title_bg_color');
    $ticket_email_divider_color       = fast_get_option('ticket_email_divider_color');
    $ticket_email_border_radius       = fast_get_option('ticket_email_border_radius');

    ob_start();
    include apply_filters('fast_filter_email_template', FAST_PLUGIN_DIR . 'templates/email/template-regular-email.php');
    $email_content = ob_get_contents();
    ob_end_clean();

    return wp_mail($to, $subject, $email_content, array(
        'Content-Type: text/html; charset="UTF-8"'
    ));
}

/**
 * Creates a slack notification content, sends it to the template and then dispatches the notification
 *
 * @param string $sender   User that triggered the action.
 * @param string $receiver User that is affected by the action.
 * @param string $message  Message to be dispatched to Slack.
 * @param array  $ticket   Ticket data.
 * @param object $comment  Comment object.
 *
 * @return array Http Header / WP_Error if the request failed.
 */
function fast_send_slack_notification($sender, $receiver, $message, $ticket, $comment)
{
    $message = apply_filters('fast_filter_replace_placeholders', $message, $ticket, $comment, $sender, $receiver);

    return wp_remote_post(fast_get_option('slack_webhook'), array(
        'method' => 'POST',
        'blocking' => false,
        'headers' => array(
            'Content-Type' => 'application/json; charset=utf-8'
        ),
        'body' => json_encode(array(
            'mrkdwn' => true,
            'text' => $message,
        ))
    ));
}

/**
 * Returns an agent's user id from his user id.
 *
 * @param int $user_id A user id.
 *
 * @return int User agent's id.
 */
function fast_get_user_agent_id($user_id)
{
    $agents_with_user_id = get_terms('fast_agent', array(
        'hide_empty' => false,
        'fields' => 'ids',
        'meta_query' => array(
            array(
                'key'       => '_fast-plugin-agent_user_id',
                'value'     => $user_id,
            )
        )
    ));

    if (!empty($agents_with_user_id) && count($agents_with_user_id) > 0) {
        return $agents_with_user_id[0];
    }
}

/**
 * Returns allowed html tags used by the plugin options
 *
 * @return array Allowed HTML tags
 */
function fast_get_allowed_html_tags()
{
    return array(
        'a' => array(
            'class' => array(),
            'href'  => array(),
            'rel'   => array(),
            'title' => array(),
        ),
        'blockquote' => array(
            'cite'  => array(),
        ),
        'div' => array(
            'class' => array(),
            'title' => array(),
            'style' => array(),
        ),
        'h1' => array(),
        'h2' => array(),
        'h3' => array(),
        'h4' => array(),
        'h5' => array(),
        'h6' => array(),
        'i' => array(),
        'li' => array(
            'class' => array(),
        ),
        'p' => array(
            'class' => array(),
        ),
        'span' => array(
            'class' => array(),
            'title' => array(),
            'style' => array(),
        ),
        'strong' => array(),
        'ul' => array(
            'class' => array(),
        ),
    );
}

/**
 * Returns number of posts found for a query of a specific user.
 *
 * @param object $filter_args Query args.
 * @param int    $is_agent is the current user an agent.
 *
 * @return int Found posts for query
 */
function fast_get_need_reply_count($filter_args, $is_agent)
{
    $filter_args['tax_query'] = isset($filter_args['tax_query']) ? $filter_args['tax_query'] : array();
    $tax_query = array_merge(
        $filter_args['tax_query'],
        array(
            array(
                'taxonomy' => 'fast_status',
                'field'    => 'slug',
                'terms'    => 'open'
            )
        )
    );

    $args = array_merge(
        $filter_args,
        array(
            'post_type' => 'fast_ticket',
            'fields' => 'ids',
            'posts_per_page' => 1,
            'meta_query' => array(
                array(
                    'key' => '_agent_replied',
                    'value' => $is_agent ? 'false' : 'true',
                )
            ),
            'tax_query' => $tax_query
        )
    );
    $query = new WP_Query($args);
    $count = $query->found_posts;
    wp_reset_postdata();
    return $count;
}

/**
 * Returns number of posts found for a query.
 *
 * @param object $filter_args Query args.
 *
 * @return int Found posts for query
 */
function fast_get_query_count($filter_args)
{
    $args = array_merge(
        $filter_args,
        array(
            'post_type' => 'fast_ticket',
            'fields' => 'ids',
            'posts_per_page' => 1,
        )
    );
    $query = new WP_Query($args);
    $count = $query->found_posts;
    wp_reset_postdata();
    return $count;
}

/**
 * Gets a users roles for checking if customer/agent/manager.
 */
function fast_get_user_roles()
{
    $user = wp_get_current_user();
    $user_roles = $user->roles;
    if (in_array('administrator', $user_roles)) {
        $user_roles[] = 'fast_agent';
        $user_roles[] = 'fast_manager';
    }
    return $user_roles;
}

/**
 * Returns a sidebar filter for each taxonomy term, along with it's query args'
 *
 * @param object $taxonomies Array of taxonomies.
 *
 * @return array Term filters
 */
function fast_create_filters_from_taxonomy_terms($taxonomies)
{
    $filters = array();
    $current_user_roles = fast_get_user_roles();

    foreach ($taxonomies as $taxonomy) {
        $tax_terms = get_terms($taxonomy, array(
            'hide_empty' => false
        ));

        foreach ($tax_terms as $term) {
            $visible_roles = fast_get_taxonomy_metabox_value($taxonomy . '_role_visibility', $term->term_id);
            $intersect = array_intersect($visible_roles, $current_user_roles);
            if (!empty($intersect)) {
                $filter = array(
                    'id' => $term->slug,
                    'name' => $term->name,
                    'args' => array(
                        'posts_per_page'=> 25,
                        'post_status' => 'publish,private',
                        'tax_query' => array(
                            array(
                                'taxonomy' => $taxonomy,
                                'field'    => 'slug',
                                'terms'    => $term->slug,
                            ),
                        ),
                    )
                );
                // Conditionally add the term count if the user is not a client.
                // For clients we call WP_Query later.
                if (!in_array('fast_customer', $current_user_roles)) {
                    $filter['args']['count'] = $term->count;
                }
                $filters[] = $filter;
            }
        }
    }
    return $filters;
}

/**
 * Returns the custom CSS to be added to the header.
 *
 * @return string CSS code.
 */
function fast_custom_css()
{
    $css                   = '';
    $register_page_image   = fast_get_option('register_page_image');
    $login_page_image      = fast_get_option('login_page_image');
    $new_ticket_page_image = fast_get_option('new_ticket_page_image');

    if (!empty($register_page_image)) {
        $css .= 'div.fast-register-image-background { background-size: cover; background-image:'
        . 'url(' . esc_url($register_page_image) . ');}';
    }
    if (!empty($login_page_image)) {
        $css .= 'div.fast-login-image-background { background-size: cover; background-image:'
        . 'url(' . esc_url($login_page_image) . ');}';
    }
    if (!empty($new_ticket_page_image)) {
        $css .= 'div.fast-new-ticket-image-background { background-size: cover; background-image: '
        . 'url(' . esc_url($new_ticket_page_image) . ');}';
    }

    return apply_filters('fast_filter_custom_css', $css);
}

/**
 * Creates oauth login url from options.
 *
 * @return string URL to envato login.
 */
function fast_generate_envato_oauth_login_url()
{
    $url = '';
    if (fast_get_option('envato_login') === 'on') {
        $support_page_id = fast_get_option('support_page');
        $envato_client_id = fast_get_option('envato_client_id');
        if (!empty($support_page_id) && !empty($envato_client_id)) {
            $redirect_url = trailingslashit(get_the_permalink($support_page_id));
            $url = 'https://api.envato.com/authorization?response_type=code&client_id=' . $envato_client_id . '&redirect_uri=' . $redirect_url;
        }
    }
    return $url;
}

/**
 * Creates oauth login url from options.
 *
 * @param string  $username Envato username.
 * @param integer $page_size Size of page to fetch.
 * @param integer $page Page to fetch.
 *
 * @return object Response object from http request to envato
 */
function fast_generate_envato_get_marketplace_items($username, $page_size, $page)
{
    return wp_remote_get('https://api.envato.com/v1/discovery/search/search/item?page=' . $page . '&page_size=' . $page_size . '&username=' . fast_get_option('envato_username'), array(
        'headers' => array(
            'timeout' => 30,
            'Authorization' => 'Bearer ' . fast_get_option('envato_token')
        )
    ));
}

/**
 * Verify envato purchase code
 *
 * @param string $purchase_code The envato purchase code.
 *
 * @return object|array Purchase details or WP_Error if fail
 */
function fast_fetch_purchase_information($purchase_code)
{
    $url = 'https://api.envato.com/v1/market/private/user/verify-purchase:' . $purchase_code . '.json';
    $token = fast_get_option('envato_token');

    $response = wp_remote_get($url, array(
        'timeout' => 30,
        'headers' => array(
            'Authorization' => 'Bearer ' . $token
        )
    ));

    if (is_wp_error($response)) {
        return $response;
    }

    $value = json_decode(wp_remote_retrieve_body($response), true);

    // Check for error responses.
    if (200 !== wp_remote_retrieve_response_code($response)) {
        return new WP_Error('cant-verify', $value['message'], array(
            'status' => 401
        ));
    }

    if (isset($value['verify-purchase']) && !empty($value['verify-purchase'])) {
        $purchase = $value['verify-purchase'];
        // Set Purchase Code.
        $purchase['purchase_code'] = $purchase_code;
        // Set ticket product.
        $products = get_terms(array(
            'hide_empty' => false,
            'taxonomy'   => 'fast_product',
            'meta_key'   => '_fast-plugin-product_envato_item',
            'meta_value' => $purchase['item_id'],
            'fields'     => 'ids'
        ));

        if (count($products) > 0) {
            $purchase['ticket_product'] = (int) $products[0];
        } else {
            return new WP_Error('fast-unknown-product', esc_html__('Could not find this product in our supported products', 'fast-plugin'), array(
                'status' => 404
            ));
        }

        return $purchase;
    } else {
        return new WP_Error('cant-verify', esc_html__('Cant Verify Purchase Code', 'fast-plugin'), array(
            'status' => 401
        ));
    }
}

/**
 * Gets a users envato purchase.
 *
 * @param string $purchase_code Envato Purchase code.
 * @param int    $user_id WP User id.
 *
 * @return boolean|WP_Post Purchase post or false if not found.
 */
function fast_get_user_envato_purchase($purchase_code, $user_id)
{
    $purchase_exists = get_posts(array(
        'post_type'      => 'fast_purchase',
        'post_status'    => 'private',
        'posts_per_page' => 1,
        'meta_key'       => '_purchase_code',
        'meta_value'     => $purchase_code,
        'author'         => $user_id
    ));

    return empty($purchase_exists) ? false : $purchase_exists[0];
}

/**
 * Gets the languages available in the app
 */
function fast_get_app_available_languages()
{
    // Get all WordPress Languages.
    require_once(ABSPATH . 'wp-admin/includes/translation-install.php');
    $wp_languages = wp_get_available_translations();

    // Get all languages in the app.
    $app_installed_languages = get_available_languages(FAST_PLUGIN_DIR . 'assets/i18n');

    // Create a new array to return.
    $app_languages = array();
    $app_languages[] = array(
        'code' => 'en_US',
        'name' => esc_html__('English (United States)', 'fast-plugin')
    );
    // Loacate the app languages in the WP array.
    foreach ($wp_languages as $language_code => $language) {
        if (in_array($language_code, $app_installed_languages)) {
            $app_languages[] = array(
                'code' => $language_code,
                'name' => $language['native_name']
            );
        }
    }
    return $app_languages;
}

/**
 * Forces a recount of all fast ticket taxonomy counts.
 * Used for 1.2.0 update to fix bug when private tickets were not being counted.
 */
function fast_recalculate_all_taxonomy_counts()
{
    $taxonomies = array(
        'fast_status',
        'fast_product',
        'fast_agent'
    );

    foreach ($taxonomies as $taxonomy) {
        $term_ids = get_terms(array(
            'taxonomy' => $taxonomy,
            'hide_empty' => false,
            'fields' => 'ids'
        ));
        wp_update_term_count_now($term_ids, $taxonomy);
    }
}

/**
 * Creates post content based on all comments made so that search works.
 *
 * @param object $ticket Ticket post object.
 *
 * @return string Post content string with concatonated comments.
 **/
function fast_create_ticket_content_from_comments($ticket)
{
    $post_content = '';
    $comments = get_comments(array(
        'post_id' => $ticket->ID
    ));
    foreach ($comments as $comment) {
        $post_content .= $comment->comment_content;
    }
    return $post_content;
}

/**
 * Adds envato products to a list
 *
 * @param array $result Product list call.
 * @param array $products List to add results to.
 *
 * @return array List of products.
 */
function fast_add_envato_products($result, $products = array())
{
    if (is_array($result['matches'])) {
        foreach ($result['matches'] as $item) {
            $products[$item['id']] = $item['name'];
        }
    }
    return $products;
}

/**
 * Gets the extra fields option.
 */
function fast_get_extra_fields()
{
    return get_option('fast_extra_fields', array(
        'fields' => array(
            array(
                'label'     => esc_html__('Site URL', 'fast-plugin'),
                'element'   => 'input',
                'inputType' => 'url',
                'id'        => 'site-url',
                'name'      => 'site-url',
                'required'  => false
            )
        )
    ));
}

/**
 * Gets the url of the frontend app.
 */
function fast_get_app_url()
{
    $url = '';
    $support_page_id = fast_get_option('support_page');
    if (!empty($support_page_id)) {
        $url = set_url_scheme(trailingslashit(get_the_permalink($support_page_id)));
    }
    return $url;
}

/**
 * Returns if woocommerce is active.
 *
 * @return boolean If woocommerce plugins is active.
 */
function fast_is_woocommerce_active()
{
    return 'woocommerce' === fast_get_option('new_ticket_type') && in_array('woocommerce/woocommerce.php', apply_filters('active_plugins', get_option('active_plugins')));
}

/**
 * Registers a user in WordPress.
 *
 * @param string $display_name User display name.
 * @param string $first_name First nam name.
 * @param string $last_name Last name.
 * @param string $email User Email.
 * @param string $password User Password.
 * @param string $language User language.
 *
 * @return object|boolean WP_User or false on failure.
 **/
function fast_register_user($display_name, $first_name, $last_name, $email, $password = null, $language = '')
{
    // Check for valid username.
    if (!validate_username($email)) {
        return new WP_Error('invalid-username', esc_html__('Username not valid', 'fast-plugin'), array(
            'status' => 400
        ));
    }
    // Check for valid email.
    if (!is_email($email)) {
        return new WP_Error('invalid-email', esc_html__('Email not valid', 'fast-plugin'), array(
            'status' => 400
        ));
    }
    // Check for conflicts.
    if (username_exists($email)) {
        return new WP_Error('username-exists', esc_html__('Username already exists', 'fast-plugin'), array(
            'status' => 409
        ));
    }
    if (email_exists($email)) {
        return new WP_Error('email-exists', esc_html__('User already registered with this email', 'fast-plugin'), array(
            'status' => 409
        ));
    }

    if (null === $password) {
        $password = wp_generate_password(10, false, false);
    }

    $user_id = wp_insert_user(array(
        'display_name' => $display_name,
        'first_name'   => $first_name,
        'last_name'    => $last_name,
        'user_login'   => $email,
        'user_pass'    => $password,
        'user_email'   => $email,
        'role'         => 'fast_customer',
        'locale'       => $language
    ));

    // Add flag to show this user has no tickets.
    add_user_meta($user_id, '_ticket_count', 0);

    // Get user object.
    $user = get_user_by('ID', $user_id);

    return $user;
}

/**
 * Encrypt and decrypt string.
 *
 * @param string $string string to be encrypted/decrypted.
 * @param string $action what to do with this? e for encrypt, d for decrypt.
 */
function fast_store_sensitive_e($string, $action = 'e')
{
    $secret_key = 'N0FAKHKJVpFpKQqnDjtFS6oKBkFhTuIVeoxLWShhxeC0Izqxgs';
    $secret_iv = 'COrUW4fb3JcHhI9XZ5vdr5UBwyPX5VFTS1Ya6ziNhmmpNawyFt';

    $encrypt_method = 'AES-256-CBC';
    $key = hash('sha256', $secret_key);
    $iv = substr(hash('sha256', $secret_iv), 0, 16);

    if ('e' === $action) {
        // Skip the $iv on versions less than 5.3.3 http://php.net/manual/en/function.openssl-encrypt.php .
        if (version_compare(PHP_VERSION, '5.3.3', '<')) {
            $string = base64_encode(openssl_encrypt($string, $encrypt_method, $key, 0));
        } else {
            $string = base64_encode(openssl_encrypt($string, $encrypt_method, $key, 0, $iv));
        }
    } elseif ('d' === $action) {
        // Skip the $iv on versions less than 5.3.3 http://php.net/manual/en/function.openssl-decrypt.php .
        if (version_compare(PHP_VERSION, '5.3.3', '<')) {
            $string = openssl_decrypt(base64_decode($string), $encrypt_method, $key, 0);
        } else {
            $string = openssl_decrypt(base64_decode($string), $encrypt_method, $key, 0, $iv);
        }
    }

    return $string;
}
